# -*- coding: utf-8 -*-

# Copyright (c) 2018-2019 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

# The source code contains parts copied and modified from Spyder project:
# https://github.com/spyder-ide/spyder
# See below for the original copyright notice.

#
# Copyright (c) Spyder Project Contributors
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import spyder
try:
    from spyder.api.plugins import SpyderPluginWidget
except ImportError:
    from spyder.plugins import SpyderPluginWidget # Spyder3

from spyder.config.base import _

from qtpy.QtWidgets import QVBoxLayout

from spyder_modelx.widgets.mxanalyzer import MxAnalyzerWidget
from .stacked_mixin import MxStackedMixin


if spyder.version_info > (5,):

    # New plugin API since Spyder 5
    from spyder.api.plugins import SpyderDockablePlugin, Plugins
    from qtpy.QtGui import QIcon
    from spyder.api.widgets.main_widget import PluginMainWidget
    # from spyder_modelx.plugins.mxplugin import ModelxConfigPage

    class MxAnalyzerMainWidget(MxStackedMixin, PluginMainWidget):

        MX_WIDGET_CLASS = MxAnalyzerWidget

        def __init__(self, name=None, plugin=None, parent=None):
            PluginMainWidget.__init__(self, name, plugin, parent)
            MxStackedMixin.__init__(self, parent=parent)

            # Layout
            layout = QVBoxLayout()
            layout.addWidget(self.stack)
            self.setLayout(layout)

            self.ipyconsole = plugin.get_plugin(Plugins.IPythonConsole)

        # --- API: methods to define or override
        # ------------------------------------------------------------------------
        def get_title(self):
            """
            Return the title that will be displayed on dockwidget or window title.
            """
            return _('MxAnalyzer')

        def setup(self):
            """
            Create widget actions, add to menu and other setup requirements.
            """
            pass

        def update_actions(self):
            """
            Update the state of exposed actions.

            Exposed actions are actions created by the self.create_action method.
            """
            pass

    class MxAnalyzerPlugin(SpyderDockablePlugin):
        """modelx sub-plugin.

        This plugin in registered by the modelx main plugin.
        """

        NAME = 'mxanalyzer'
        WIDGET_CLASS = MxAnalyzerMainWidget
        REQUIRES = [Plugins.IPythonConsole, 'modelx_plugin']
        CONF_SECTION = 'modelx'
        CONF_FILE = False

        # -------------------------------------------------------------------
        # --- API: Mandatory methods to define ------------------------------

        def get_name(self):
            """
            Return the plugin localized name.

            Returns
            -------
            str
                Localized name of the plugin.

            Notes
            -----
            This is a method to be able to update localization without a restart.
            """
            return _('MxAnalyzer')

        def get_description(self):
            """
            Return the plugin localized description.

            Returns
            -------
            str
                Localized description of the plugin.

            Notes
            -----
            This is a method to be able to update localization without a restart.
            """
            return _('Widget for tracing modelx node dependency')

        def get_icon(self):
            """
            Return the plugin associated icon.

            Returns
            -------
            QIcon
                QIcon instance
            """
            return QIcon()

        def register(self):
            """
            Setup and register plugin in Spyder's main window and connect it to
            other plugins.
            """
            self.get_plugin('modelx_plugin').get_container().set_child_plugin('analyzer', self.get_container())
        # -------------------------------------------------------------------

else:
    class MxAnalyzerPlugin(MxStackedMixin, SpyderPluginWidget):
        """modelx sub-plugin.

        This plugin in registered by the modelx main plugin.
        """

        CONF_SECTION = 'modelx_analyzer'
        MX_WIDGET_CLASS = MxAnalyzerWidget
        CONF_FILE = False

        def __init__(self, parent=None):

            SpyderPluginWidget.__init__(self, parent)
            MxStackedMixin.__init__(self, parent)

            # Layout
            layout = QVBoxLayout()
            layout.addWidget(self.stack)
            if spyder.version_info > (4,):
                self.options_button.setVisible(False)
            self.setLayout(layout)

            if spyder.version_info < (4,):
                # Initialize plugin
                self.initialize_plugin()

        # --- SpyderPluginWidget API ----------------------------------------------
        def get_plugin_title(self):
            """Return widget title."""
            return 'MxAnalyzer'

        def get_focus_widget(self):
            """Return the widget to give focus to."""
            return self.current_widget()

        def refresh_plugin(self):
            """Refresh MxExplorer widget."""
            pass

        def get_plugin_actions(self):
            """Return a list of actions related to plugin."""
            return []

        def register_plugin(self):
            """Register plugin in Spyder's main window."""
            if spyder.version_info < (4,):
                self.main.add_dockwidget(self)
            else:
                self.add_dockwidget()

        def on_first_registration(self):
            """Action to be performed on first plugin registration."""
            self.main.tabify_plugins(self.main.help, self)

        def apply_plugin_settings(self, options):
            """Apply configuration file's plugin settings."""
            pass



# -*- coding: utf-8 -*-

# Copyright (c) 2018-2019 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

# The source code contains parts copied and modified from Spyder project:
# https://github.com/spyder-ide/spyder
# See below for the original copyright notice.

#
# Copyright (c) Spyder Project Contributors
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.


"""modelx Plugin."""
from jupyter_core.paths import jupyter_config_dir, jupyter_runtime_dir

try:
    from spyder.api.plugins import SpyderPluginWidget
except ImportError:
    from spyder.plugins import SpyderPluginWidget # Spyder3
try:
    from spyder.api.preferences import PluginConfigPage
except ImportError:
    from spyder.plugins.configdialog import PluginConfigPage # Spyder3

from qtpy.QtCore import Qt, Signal, Slot
from qtpy.QtWidgets import (QVBoxLayout, QHBoxLayout, QLabel,
                            QSplitter, QStackedWidget)
import spyder
from spyder.config.base import (_, DEV, get_conf_path, get_home_dir,
                                get_module_path)
from spyder.utils import icon_manager as ima
from spyder.py3compat import is_string, PY2, to_text_string
if spyder.version_info < (4,):
    from spyder.config.main import CONF # Spyder 3
else:
    from spyder.config.manager import CONF
from spyder.utils import encoding, programs, sourcecode
from spyder.utils.qthelpers import (add_actions, create_action,
                                    create_toolbutton, create_plugin_layout)
if spyder.version_info < (4,):
    from spyder.plugins.ipythonconsole import IPythonConsole
else:
    from spyder.plugins.ipythonconsole.plugin import IPythonConsole

from spyder_modelx.mxkernelspec import MxKernelSpec
from spyder_modelx.widgets.mxexplorer import MxMainWidget
from spyder_modelx.widgets.mxclient import MxClientWidget

from .stacked_mixin import MxStackedMixin
from .dataview_plugin import MxDataViewPlugin
from .analyzer_plugin import MxAnalyzerPlugin


class ModelxConfigPage(PluginConfigPage):
    """modelx plugin preferences."""

    def get_name(self):
        return _('modelx')

    def setup_page(self):
        pass


# TODO: Split plugins into separate files and place them under plugin folder.

if spyder.version_info > (5,):

    # New plugin API since Spyder 5
    from spyder.plugins.mainmenu.api import (
        ApplicationMenus, ConsolesMenuSections, HelpMenuSections)
    from spyder.api.plugins import SpyderDockablePlugin, Plugins
    from qtpy.QtGui import QIcon
    from spyder.api.widgets.main_widget import PluginMainWidget

    class MxPluginMainWidgetActions:

        OpenNewConsole = 'new_console'

    class MxPluginMainWidgetActionsOptionsMenuSections:
        Main = 'main_section'

    class MxPluginMainWidget(MxStackedMixin, PluginMainWidget):

        MX_WIDGET_CLASS = MxMainWidget

        def __init__(self, name=None, plugin=None, parent=None):
            PluginMainWidget.__init__(self, name, plugin, parent)
            MxStackedMixin.__init__(self, parent=parent)

            # Layout
            layout = QVBoxLayout()
            layout.addWidget(self.stack)
            self.setLayout(layout)

            self.ipyconsole = plugin.get_plugin(Plugins.IPythonConsole)

            # To avoid circular dependency,
            # Set by MxAnalyzer and MxDataViewer's register method
            # calling set_child_plugin
            self.analyzer = None
            self.dataview = None

        def set_child_plugin(self, attr, widget):
            setattr(self, attr, widget)

        # --- API: methods to define or override
        # ------------------------------------------------------------------------
        def get_title(self):
            """
            Return the title that will be displayed on dockwidget or window title.
            """
            return _('MxExplorer')

        def setup(self):
            """
            Create widget actions, add to menu and other setup requirements.
            """

            # ---- Toolbar actions
            create_client_action = new_console_action = self.create_action(
                MxPluginMainWidgetActions.OpenNewConsole,
                text=_('New MxConsole'),
                icon=self.create_icon('ipython_console'),
                triggered=self.create_new_client
            )

            # Options menu
            options_menu = self.get_options_menu()
            for item in [new_console_action]:
                self.add_item_to_menu(
                    item,
                    menu=options_menu,
                    section=MxPluginMainWidgetActionsOptionsMenuSections.Main
                )

            if spyder.version_info > (5, 0, 3):
                for console_new_action in [create_client_action]:
                    self.main.mainmenu.add_item_to_application_menu(
                        console_new_action,
                        menu_id=ApplicationMenus.Consoles,
                        section=ConsolesMenuSections.New)

                # Plugin actions
                self.menu_actions = [create_client_action] + \
                                    self.ipyconsole.menu_actions.copy()

                self.ipyconsole.menu_actions.insert(
                    0, create_client_action
                )
            else:
                # Add the action to the 'Consoles' menu on the main window
                main_consoles_menu = self.main.consoles_menu_actions
                main_consoles_menu.insert(0, create_client_action)
                self.main.ipyconsole.menu_actions.insert(0, create_client_action)

                # Plugin actions
                self.menu_actions = [create_client_action] + \
                                    self.main.ipyconsole.menu_actions.copy()

                add_actions(self.main.ipyconsole.tabwidget.menu,
                            [create_client_action],
                            insert_before=main_consoles_menu[1])

        def update_actions(self):
            """
            Update the state of exposed actions.

            Exposed actions are actions created by the self.create_action method.
            """
            pass

        # --- IPython Console methods
        # ------------------------------------------------------------------------
        @Slot()
        @Slot(bool)
        @Slot(str)
        @Slot(bool, str)
        @Slot(bool, bool)
        @Slot(bool, str, bool)
        def create_new_client(self, give_focus=True, filename='', is_cython=False,
                              **kwargs):
            """Create a new client

            Copied and modified from spyder.plugins.ipythonconsole.IPythonConsole
            """
            ipycon = self.ipyconsole

            ipycon.master_clients += 1
            client_id = dict(int_id=to_text_string(ipycon.master_clients),
                             str_id='A')
            cf = ipycon._new_connection_file()
            show_elapsed_time = ipycon.get_option('show_elapsed_time')
            reset_warning = ipycon.get_option('show_reset_namespace_warning')
            if spyder.version_info > (5,):
                client_kwargs = {
                    "ask_before_restart": ipycon.get_option('ask_before_restart'),
                    "ask_before_closing": ipycon.get_option('ask_before_closing'),
                    "options_button": ipycon.options_button,
                    "css_path": ipycon.css_path
                }
                if "given_name" in kwargs:  # if not use default 'MxConsole'
                    client_kwargs["given_name"] = kwargs["given_name"]
            elif spyder.version_info > (4,):
                client_kwargs = {
                    "ask_before_restart": ipycon.get_option('ask_before_restart'),
                    "options_button": ipycon.options_button,
                    "css_path": ipycon.css_path
                }
                if spyder.version_info > (4, 2, 0):
                    client_kwargs["ask_before_closing"] = ipycon.get_option('ask_before_closing')
                if "given_name" in kwargs:
                    client_kwargs["given_name"] = kwargs["given_name"]
            elif spyder.version_info > (3, 3, 5):
                client_kwargs = {
                    "ask_before_restart": ipycon.get_option('ask_before_restart')}
            else:
                client_kwargs = {}

            addops = {}
            if "is_pylab" in kwargs:
                addops["is_pylab"] = kwargs["is_pylab"]
            if "is_sympy" in kwargs:
                addops["is_sympy"] = kwargs["is_sympy"]

            client = MxClientWidget(ipycon,
                                    id_=client_id,
                                    history_filename=get_conf_path('history.py'),
                                    config_options=ipycon.config_options(),
                                    additional_options=ipycon.additional_options(**addops),
                                    interpreter_versions=ipycon.interpreter_versions(),
                                    connection_file=cf,
                                    menu_actions=self.menu_actions,
                                    show_elapsed_time=show_elapsed_time,
                                    reset_warning=reset_warning,
                                    **client_kwargs)

            # Change stderr_dir if requested
            if spyder.version_info < (3, 3, 2):
                testing = ipycon.testing
            else:
                testing = (ipycon.test_dir is not None)
            if testing:
                client.stderr_dir = ipycon.test_dir

            ipycon.add_tab(client, name=client.get_name(), filename=filename)

            if cf is None:
                error_msg = ipycon.permission_error_msg.format(jupyter_runtime_dir())
                client.show_kernel_error(error_msg)
                return

            # Check if ipykernel is present in the external interpreter.
            # Else we won't be able to create a client
            if not CONF.get('main_interpreter', 'default'):
                pyexec = CONF.get('main_interpreter', 'executable')
                if spyder.version_info > (3, 2, 8):
                    has_ipykernel = programs.is_module_installed(
                        "spyder_kernels",
                        interpreter=pyexec)     # missing version param
                    testcond = has_ipykernel
                else:
                    has_ipykernel = programs.is_module_installed(
                        "ipykernel",
                        interpreter=pyexec)
                    has_cloudpickle = programs.is_module_installed(
                        'cloudpickle',
                        interpreter=pyexec)     # this is gone now
                    testcond = (has_ipykernel and has_cloudpickle)

                if not testcond:
                    client.show_kernel_error(_("Your Python environment or "
                                               "installation doesn't "
                                               "have the <tt>ipykernel</tt> and "
                                               "<tt>cloudpickle</tt> modules "
                                               "installed on it. Without these modules "
                                               "is not possible for Spyder to create a "
                                               "console for you.<br><br>"
                                               "You can install them by running "
                                               "in a system terminal:<br><br>"
                                               "<tt>pip install ipykernel cloudpickle</tt>"
                                               "<br><br>"
                                               "or<br><br>"
                                               "<tt>conda install ipykernel cloudpickle</tt>"))
                    return

            self.connect_client_to_kernel(client, is_cython=is_cython)
            if client.shellwidget.kernel_manager is None:
                return
            ipycon.register_client(client)

        def connect_client_to_kernel(self, client, is_cython=False,
                                     **kwargs):    # kwargs for is_pylab, is_sympy
            """Connect a client to its kernel

            Copied and modified from spyder.plugins.ipythonconsole.IPythonConsole
            """
            ipycon = self.main.ipyconsole

            connection_file = client.connection_file

            if ipycon.test_no_stderr:
                stderr_handle = None
            else:
                if spyder.version_info < (3, 3, 2):
                    stderr_handle = client.stderr_file
                else:
                    stderr_handle = client.stderr_handle

            km, kc = self.create_kernel_manager_and_kernel_client(
                connection_file,
                stderr_handle,
                is_cython=is_cython, **kwargs)

            # An error occurred if this is True
            if is_string(km) and kc is None:
                client.shellwidget.kernel_manager = None
                client.show_kernel_error(km)
                return

            # This avoids a recurrent, spurious NameError when running our
            # tests in our CIs
            if not ipycon.testing:
                kc.started_channels.connect(
                    lambda c=client: self.process_started(c))
                kc.stopped_channels.connect(
                    lambda c=client: self.process_finished(c))
            kc.start_channels(shell=True, iopub=True)

            shellwidget = client.shellwidget
            if spyder.version_info > (5,):
                shellwidget.set_kernel_client_and_manager(kc, km)
                shellwidget.sig_exception_occurred.connect(
                    self.main.console.handle_exception)
            elif spyder.version_info > (4,):
                shellwidget.set_kernel_client_and_manager(kc, km)
                shellwidget.sig_exception_occurred.connect(
                    self.main.console.exception_occurred)
            else:
                shellwidget.kernel_manager = km
                shellwidget.kernel_client = kc

        def create_kernel_spec(self, is_cython=False, **kwargs):
            """Create a kernel spec for our own kernels

            Copied and modified from spyder.plugins.ipythonconsole.IPythonConsole
            """
            ipycon = self.main.ipyconsole
            # Before creating our kernel spec, we always need to
            # set this value in spyder.ini
            if spyder.version_info > (3, 3, 1) or not ipycon.testing:
                CONF.set('main', 'spyder_pythonpath',
                         ipycon.main.get_spyder_pythonpath())
            return MxKernelSpec(is_cython=is_cython, **kwargs)

        def create_kernel_manager_and_kernel_client(self, connection_file,
                                                    stderr_handle,
                                                    is_cython=False):

            if spyder.version_info < (3, 2, 8):
                return IPythonConsole.create_kernel_manager_and_kernel_client(
                    self, connection_file, stderr_handle)
            else:
                return IPythonConsole.create_kernel_manager_and_kernel_client(
                    self, connection_file, stderr_handle,
                    is_cython=is_cython)

        def process_started(self, client):
            self.main.ipyconsole.process_started(client)
            self.add_shellwidget(client.shellwidget)
            if self.analyzer is not None:
                self.analyzer.add_shellwidget(client.shellwidget)
            if self.dataview is not None:
                self.dataview.add_shellwidget(client.shellwidget)

        def process_finished(self, client):
            self.main.ipyconsole.process_finished(client)
            self.remove_shellwidget(id(client.shellwidget))
            if self.analyzer is not None:
                self.analyzer.remove_shellwidget(id(client.shellwidget))
            if self.dataview is not None:
                self.dataview.remove_shellwidget(id(client.shellwidget))

    class ModelxPlugin(SpyderDockablePlugin):
        """modelx plugin."""

        NAME = 'modelx_plugin'
        WIDGET_CLASS = MxPluginMainWidget
        REQUIRES = [Plugins.IPythonConsole]
        CONF_SECTION = 'modelx'
        CONFIGWIDGET_CLASS = ModelxConfigPage
        CONF_FILE = False

        # -------------------------------------------------------------------
        # --- API: Mandatory methods to define ------------------------------

        def get_name(self):
            """
            Return the plugin localized name.

            Returns
            -------
            str
                Localized name of the plugin.

            Notes
            -----
            This is a method to be able to update localization without a restart.
            """
            return _('MxExplorer')

        def get_description(self):
            """
            Return the plugin localized description.

            Returns
            -------
            str
                Localized description of the plugin.

            Notes
            -----
            This is a method to be able to update localization without a restart.
            """
            return _('Main widget of the plugin for modelx')

        def get_icon(self):
            """
            Return the plugin associated icon.

            Returns
            -------
            QIcon
                QIcon instance
            """
            return QIcon()

        def register(self):
            """
            Setup and register plugin in Spyder's main window and connect it to
            other plugins.
            """
            self.register_subplugin()

        # -------------------------------------------------------------------

        def register_subplugin(self):
            """Register sub plugin """
            pass

else:
    class ModelxPlugin(MxStackedMixin, SpyderPluginWidget):
        """modelx plugin."""

        CONF_SECTION = 'modelx'
        CONFIGWIDGET_CLASS = ModelxConfigPage
        MX_WIDGET_CLASS = MxMainWidget
        CONF_FILE = False

        def __init__(self, parent=None, testing=False):
            SpyderPluginWidget.__init__(self, parent)
            MxStackedMixin.__init__(self, parent)

            # Layout
            layout = QVBoxLayout()
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(self.stack)
            if spyder.version_info > (4,):
                self.options_button.setVisible(False)
            self.setLayout(layout)

            if spyder.version_info < (4,):
                # Initialize plugin
                if not testing:
                    self.initialize_plugin()

        # --- SpyderPluginWidget API ----------------------------------------------
        def get_plugin_title(self):
            """Return widget title."""
            return _('MxExplorer')

        def get_focus_widget(self):
            """Return the widget to give focus to."""
            return self.current_widget()

        def refresh_plugin(self):
            """Refresh MxExplorer widget."""
            pass

        def get_plugin_actions(self):
            """Return a list of actions related to plugin.

            Note: these actions will be enabled when plugin's dockwidget is visible
                  and they will be disabled when it's hidden
            """

            create_client_action = create_action(
                self,
                _("New MxConsole"),
                icon=ima.icon('ipython_console'),
                triggered=self.create_new_client,
                context=Qt.WidgetWithChildrenShortcut)
            self.register_shortcut(create_client_action, context="ipython_console",
                                   name="New tab")

            # Add the action to the 'Consoles' menu on the main window
            main_consoles_menu = self.main.consoles_menu_actions
            main_consoles_menu.insert(0, create_client_action)
            self.main.ipyconsole.menu_actions.insert(0, create_client_action)

            # Plugin actions
            self.menu_actions = [create_client_action] + \
                                self.main.ipyconsole.menu_actions.copy()

            add_actions(self.main.ipyconsole.tabwidget.menu,
                        [create_client_action],
                        insert_before=main_consoles_menu[1])

            # This should return the actions specific to this plugin.
            return [create_client_action]

        def register_plugin(self):
            """Register plugin in Spyder's main window."""
            if spyder.version_info < (4,):
                self.main.add_dockwidget(self)
            else:
                self.add_dockwidget()
            self.register_subplugin()

        def register_subplugin(self):
            """Register sub plugin """
            self.dataview = MxDataViewPlugin(self.main)
            self.main.thirdparty_plugins.append(self.dataview)
            self.dataview.register_plugin()

            self.analyzer = MxAnalyzerPlugin(self.main)
            self.main.thirdparty_plugins.append(self.analyzer)
            self.analyzer.register_plugin()

        def on_first_registration(self):
            """Action to be performed on first plugin registration."""
            self.main.tabify_plugins(self.main.help, self)

        def apply_plugin_settings(self, options):
            """Apply configuration file's plugin settings."""
            pass

        @Slot()
        @Slot(bool)
        @Slot(str)
        @Slot(bool, str)
        @Slot(bool, bool)
        @Slot(bool, str, bool)
        def create_new_client(self, give_focus=True, filename='', is_cython=False,
                              **kwargs):
            """Create a new client

            Copied and modified from spyder.plugins.ipythonconsole.IPythonConsole
            """
            ipycon = self.main.ipyconsole

            ipycon.master_clients += 1
            client_id = dict(int_id=to_text_string(ipycon.master_clients),
                             str_id='A')
            cf = ipycon._new_connection_file()
            show_elapsed_time = ipycon.get_option('show_elapsed_time')
            reset_warning = ipycon.get_option('show_reset_namespace_warning')
            if spyder.version_info > (5,):
                client_kwargs = {
                    "ask_before_restart": ipycon.get_option('ask_before_restart'),
                    "ask_before_closing": ipycon.get_option('ask_before_closing'),
                    "options_button": ipycon.options_button,
                    "css_path": ipycon.css_path
                }
                if "given_name" in kwargs:  # if not use default 'MxConsole'
                    client_kwargs["given_name"] = kwargs["given_name"]
            elif spyder.version_info > (4,):
                client_kwargs = {
                    "ask_before_restart": ipycon.get_option('ask_before_restart'),
                    "options_button": ipycon.options_button,
                    "css_path": ipycon.css_path
                }
                if spyder.version_info > (4, 2, 0):
                    client_kwargs["ask_before_closing"] = ipycon.get_option('ask_before_closing')
                if "given_name" in kwargs:
                    client_kwargs["given_name"] = kwargs["given_name"]
            elif spyder.version_info > (3, 3, 5):
                client_kwargs = {
                    "ask_before_restart": ipycon.get_option('ask_before_restart')}
            else:
                client_kwargs = {}

            addops = {}
            if "is_pylab" in kwargs:
                addops["is_pylab"] = kwargs["is_pylab"]
            if "is_sympy" in kwargs:
                addops["is_sympy"] = kwargs["is_sympy"]

            client = MxClientWidget(ipycon,
                                    id_=client_id,
                                    history_filename=get_conf_path('history.py'),
                                    config_options=ipycon.config_options(),
                                    additional_options=ipycon.additional_options(**addops),
                                    interpreter_versions=ipycon.interpreter_versions(),
                                    connection_file=cf,
                                    menu_actions=self.menu_actions,
                                    show_elapsed_time=show_elapsed_time,
                                    reset_warning=reset_warning,
                                    **client_kwargs)

            # Change stderr_dir if requested
            if spyder.version_info < (3, 3, 2):
                testing = ipycon.testing
            else:
                testing = (ipycon.test_dir is not None)
            if testing:
                client.stderr_dir = ipycon.test_dir

            ipycon.add_tab(client, name=client.get_name(), filename=filename)

            if cf is None:
                error_msg = ipycon.permission_error_msg.format(jupyter_runtime_dir())
                client.show_kernel_error(error_msg)
                return

            # Check if ipykernel is present in the external interpreter.
            # Else we won't be able to create a client
            if not CONF.get('main_interpreter', 'default'):
                pyexec = CONF.get('main_interpreter', 'executable')
                if spyder.version_info > (3, 2, 8):
                    has_ipykernel = programs.is_module_installed(
                        "spyder_kernels",
                        interpreter=pyexec)     # missing version param
                    testcond = has_ipykernel
                else:
                    has_ipykernel = programs.is_module_installed(
                        "ipykernel",
                        interpreter=pyexec)
                    has_cloudpickle = programs.is_module_installed(
                        'cloudpickle',
                        interpreter=pyexec)     # this is gone now
                    testcond = (has_ipykernel and has_cloudpickle)

                if not testcond:
                    client.show_kernel_error(_("Your Python environment or "
                                               "installation doesn't "
                                               "have the <tt>ipykernel</tt> and "
                                               "<tt>cloudpickle</tt> modules "
                                               "installed on it. Without these modules "
                                               "is not possible for Spyder to create a "
                                               "console for you.<br><br>"
                                               "You can install them by running "
                                               "in a system terminal:<br><br>"
                                               "<tt>pip install ipykernel cloudpickle</tt>"
                                               "<br><br>"
                                               "or<br><br>"
                                               "<tt>conda install ipykernel cloudpickle</tt>"))
                    return

            self.connect_client_to_kernel(client, is_cython=is_cython)
            if client.shellwidget.kernel_manager is None:
                return
            ipycon.register_client(client)

        def connect_client_to_kernel(self, client, is_cython=False,
                                     **kwargs):    # kwargs for is_pylab, is_sympy
            """Connect a client to its kernel

            Copied and modified from spyder.plugins.ipythonconsole.IPythonConsole
            """
            ipycon = self.main.ipyconsole

            connection_file = client.connection_file

            if ipycon.test_no_stderr:
                stderr_handle = None
            else:
                if spyder.version_info < (3, 3, 2):
                    stderr_handle = client.stderr_file
                else:
                    stderr_handle = client.stderr_handle

            km, kc = self.create_kernel_manager_and_kernel_client(
                connection_file,
                stderr_handle,
                is_cython=is_cython, **kwargs)

            # An error occurred if this is True
            if is_string(km) and kc is None:
                client.shellwidget.kernel_manager = None
                client.show_kernel_error(km)
                return

            # This avoids a recurrent, spurious NameError when running our
            # tests in our CIs
            if not ipycon.testing:
                kc.started_channels.connect(
                    lambda c=client: self.process_started(c))
                kc.stopped_channels.connect(
                    lambda c=client: self.process_finished(c))
            kc.start_channels(shell=True, iopub=True)

            shellwidget = client.shellwidget
            if spyder.version_info > (5,):
                shellwidget.set_kernel_client_and_manager(kc, km)
                shellwidget.sig_exception_occurred.connect(
                    self.main.console.handle_exception)
            elif spyder.version_info > (4,):
                shellwidget.set_kernel_client_and_manager(kc, km)
                shellwidget.sig_exception_occurred.connect(
                    self.main.console.exception_occurred)
            else:
                shellwidget.kernel_manager = km
                shellwidget.kernel_client = kc

        def create_kernel_spec(self, is_cython=False, **kwargs):
            """Create a kernel spec for our own kernels

            Copied and modified from spyder.plugins.ipythonconsole.IPythonConsole
            """
            ipycon = self.main.ipyconsole
            # Before creating our kernel spec, we always need to
            # set this value in spyder.ini
            if spyder.version_info > (3, 3, 1) or not ipycon.testing:
                CONF.set('main', 'spyder_pythonpath',
                         ipycon.main.get_spyder_pythonpath())
            return MxKernelSpec(is_cython=is_cython, **kwargs)

        def create_kernel_manager_and_kernel_client(self, connection_file,
                                                    stderr_handle,
                                                    is_cython=False):

            if spyder.version_info < (3, 2, 8):
                return IPythonConsole.create_kernel_manager_and_kernel_client(
                    self, connection_file, stderr_handle)
            else:
                return IPythonConsole.create_kernel_manager_and_kernel_client(
                    self, connection_file, stderr_handle,
                    is_cython=is_cython)

        def process_started(self, client):
            self.main.ipyconsole.process_started(client)
            self.add_shellwidget(client.shellwidget)
            if self.analyzer is not None:
                self.analyzer.add_shellwidget(client.shellwidget)
            if self.dataview is not None:
                self.dataview.add_shellwidget(client.shellwidget)

        def process_finished(self, client):
            self.main.ipyconsole.process_finished(client)
            self.remove_shellwidget(id(client.shellwidget))
            if self.analyzer is not None:
                self.analyzer.remove_shellwidget(id(client.shellwidget))
            if self.dataview is not None:
                self.dataview.remove_shellwidget(id(client.shellwidget))








# -*- coding: utf-8 -*-

# Copyright (c) 2018-2019 Fumito Hamamura <fumito.ham@gmail.com>

# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/>.

# The source code contains parts copied and modified from Spyder project:
# https://github.com/spyder-ide/spyder
# See below for the original copyright notice.

#
# Copyright (c) Spyder Project Contributors
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import sys
import ast

from qtpy.QtWidgets import (
    QVBoxLayout, QWidget, QLabel, QButtonGroup, QRadioButton, QGridLayout,
    QHBoxLayout, QPushButton
)

import spyder
try:
    from spyder.api.plugins import SpyderPluginWidget
except ImportError:
    from spyder.plugins import SpyderPluginWidget  # Spyder3

from spyder.utils.qthelpers import create_plugin_layout, create_action

if spyder.version_info > (5,):
    from spyder_modelx.widgets.mxdataviewer.dataframeviewer import MxDataFrameViewer
    from spyder_modelx.widgets.mxdataviewer.arrayviewer import MxArrayViewer
    from spyder_modelx.widgets.mxdataviewer.collectionsviewer import MxCollectionsViewer
elif spyder.version_info > (4,):
    from spyder_modelx.widgets.mxdataviewer.compat401.dataframeviewer import MxDataFrameViewer
    from spyder_modelx.widgets.mxdataviewer.compat401.arrayviewer import MxArrayViewer
    from spyder_modelx.widgets.mxdataviewer.compat401.collectionsviewer import MxCollectionsViewer
else:
    from spyder_modelx.widgets.mxdataviewer.compat32.dataframeviewer import MxDataFrameViewer
    from spyder_modelx.widgets.mxdataviewer.compat32.arrayviewer import MxArrayViewer
    if spyder.version_info > (3, 3):
        from spyder_modelx.widgets.mxdataviewer.compat33.collectionsviewer import MxCollectionsViewer
    else:
        from spyder_modelx.widgets.mxdataviewer.compat32.collectionsviewer import MxCollectionsViewer

from spyder_modelx.widgets.mxtoolbar import MxToolBarMixin
from spyder_modelx.widgets.mxlineedit import MxPyExprLineEdit
from spyder.config.base import _
from .stacked_mixin import MxStackedMixin


if spyder.version_info > (5,):

    class MxDataViewWidget(QWidget):

        def __init__(self, parent, **kwargs):
            QWidget.__init__(self, parent)

            self.plugin = parent.get_plugin()
            # self.plugin_actions = []

            #         outer_layout
            #        +----------------------------------------------|-----------+
            #        |   inner_layout                               |           |
            #        |  +---------|-----------------------------+   |           |
            #  upper |  |         | +----------------------+    |   |           |
            # layout |  |obj_radio| |objbox_layout         |    |   |           |
            #        |  |         | +----------------------+    |   |  update   |
            #        |  -----------------------------------------   |  button   |
            #        |  |   expr  |                             |   |           |
            #        |  |   radio |  exprbox                    |   |           |
            #        |  +---------|-----------------------------+   |           |
            #        |                                              |           |
            #        -----------------------------------------------|------------
            #        |                                                          |
            #        |   msgbox                                                 |
            #        |                                                          |
            #        +----------------------------------------------------------+
            #

            button_group = QButtonGroup(parent=self)
            self.object_radio = object_radio = QRadioButton("Object")
            self.expr_radio = expr_radio = QRadioButton("Expression")
            button_group.addButton(object_radio)
            button_group.addButton(expr_radio)

            object_radio.toggled.connect(self.activateObject)
            expr_radio.toggled.connect(self.activateExpression)

            update_button = QPushButton(text="Update", parent=self)
            update_button.clicked.connect(self.update_data)

            # txt = _("Expression")
            # if sys.platform == 'darwin':
            #     expr_label = QLabel("  " + txt)
            # else:
            #     expr_label = QLabel(txt)

            font = self.plugin.get_font()

            self.exprbox = MxPyExprLineEdit(self, font=font)
            self.objbox = QLabel(parent=self)
            self.argbox = MxPyExprLineEdit(self, font=font)
            self.msgbox = QLabel(parent=self)
            self.msgbox.setText("")
            self.msgbox.setWordWrap(True)

            outer_layout = QVBoxLayout()
            upper_layout = QHBoxLayout()
            outer_layout.addLayout(upper_layout)
            outer_layout.addWidget(self.msgbox)

            inner_layout = QGridLayout()
            inner_layout.addWidget(object_radio, 0, 0)
            inner_layout.addWidget(expr_radio, 1, 0)
            inner_layout.addWidget(self.exprbox, 1, 1)
            objbox_layout = QHBoxLayout()
            objbox_layout.addWidget(self.objbox)
            objbox_layout.addWidget(self.argbox)
            objbox_layout.setStretch(0, 3)  # 3:1
            objbox_layout.setStretch(1, 1)
            inner_layout.addLayout(objbox_layout, 0, 1)

            upper_layout.addLayout(inner_layout)
            upper_layout.addWidget(update_button)

            # widget = QWidget(parent=self)
            # widget.setLayout(outer_layout)

            # Create main widget
            self.widget = QWidget(parent=self) # MxDataFrameViewer(self)

            self.main_layout = main_layout = QVBoxLayout()
            main_layout.addLayout(outer_layout)
            main_layout.addWidget(self.widget)
            main_layout.setStretch(1, 1)

            # Main layout of this widget
            # layout = create_plugin_layout(self.tools_layout)
            # layout.addLayout(main_layout)

            margins = (0, 0, 0, 0)

            for lo in [outer_layout, upper_layout, inner_layout, self.msgbox]:
                lo.setContentsMargins(*margins)

            objbox_layout.setContentsMargins(0, 0, 0, 0)

            self.setLayout(main_layout)

            self.attrdict = None
            object_radio.setChecked(True)

            # self.calc_on_update = create_action(self,
            #     _("Calculate upon update"),
            #     tip=_("Calculate the selected cells if not yet calculated"))
            # self.calc_on_update.setChecked(True)
            #
            # self.actions = [self.calc_on_update]

        def set_shellwidget(self, shellwidget):
            """Bind shellwidget instance to namespace browser"""
            self.shellwidget = shellwidget
            self.shellwidget.set_mxdataview(self)

        def activateObject(self, checked):
            self.argbox.setEnabled(checked)
            self.exprbox.setEnabled(not checked)

        def activateExpression(self, checked):
            self.argbox.setEnabled(not checked)
            self.exprbox.setEnabled(checked)

        def update_object(self, data):
            if data is None:
                return

            self.attrdict = data
            if data["type"] == 'Reference' or (
                    data["type"] == "Cells" and not data["parameters"]):
                self.argbox.setText("")
                self.argbox.setEnabled(False)
            else:
                self.argbox.setEnabled(True)

            self.objbox.setText(data['_evalrepr'])

        def update_data(self):
            if self.object_radio.isChecked():

                argtxt = self.argbox.get_expr()
                args = "(" + argtxt + ("," if argtxt else "") + ")"
                # assert
                ast.literal_eval(args)

                calc = self.plugin.get_container().calc_on_update_action.isChecked()

                val, is_calculated = self.shellwidget.update_mxdataview(
                    is_obj=True,
                    obj=self.attrdict["fullname"],
                    args=args,
                    calc=calc
                )
                self.update_value(val)
                if is_calculated:
                    self.shellwidget.refresh_namespacebrowser()

            elif self.expr_radio.isChecked():
                self.shellwidget.update_mxdataview(
                    is_obj=False,
                    expr=self.exprbox.get_expr()
                )
                self.shellwidget.refresh_namespacebrowser()
            else:
                raise RuntimeError("MxDataViewer: must not happen")

        def update_value(self, data):

            import pandas as pd
            import numpy as np
            import numpy.ma

            if isinstance(data, (pd.DataFrame, pd.Index, pd.Series)):
                self.widget.deleteLater()
                self.widget = MxDataFrameViewer(self)
                self.widget.setup_and_check(data)
                self.msgbox.setText(data.__class__.__name__)
            elif isinstance(data, (np.ndarray, np.ma.MaskedArray)):
                self.widget.deleteLater()
                self.widget = MxArrayViewer(self)
                self.widget.setup_and_check(data, title='', readonly=True)
                self.msgbox.setText(data.__class__.__name__)
            elif isinstance(data, (list, set, tuple, dict)):
                self.widget.deleteLater()
                self.widget = MxCollectionsViewer(self)
                self.widget.setup(data, title='', readonly=True)
                self.msgbox.setText(data.__class__.__name__)
            else:
                txt = data.__class__.__name__ + ": " + repr(data)
                self.widget.deleteLater()
                self.widget = QWidget(parent=self)
                self.msgbox.setText(txt)

            self.main_layout.addWidget(self.widget)
            self.main_layout.setStretchFactor(self.widget, 1)

else:

    class MxDataViewWidget(MxToolBarMixin, QWidget):

        def __init__(self, parent, **kwargs):
            QWidget.__init__(self, parent)

            if spyder.version_info > (5,):
                self.plugin = parent.get_plugin()
            else:
                self.plugin = parent

            # Create tool bar
            if "options_button" in kwargs:
                self.options_button = kwargs["options_button"]
            else:
                self.options_button = None
            self.plugin_actions = []
            MxToolBarMixin.__init__(
                self,
                options_button=self.options_button,
                plugin_actions=self.plugin_actions
            )

            #
            #         outer_layout
            #        +----------------------------------------------|-----------+
            #        |   inner_layout                               |           |
            #        |  +---------|-----------------------------+   |           |
            #  upper |  |         | +----------------------+    |   |           |
            # layout |  |obj_radio| |objbox_layout         |    |   |           |
            #        |  |         | +----------------------+    |   |  update   |
            #        |  -----------------------------------------   |  button   |
            #        |  |   expr  |                             |   |           |
            #        |  |   radio |  exprbox                    |   |           |
            #        |  +---------|-----------------------------+   |           |
            #        |                                              |           |
            #        -----------------------------------------------|------------
            #        |                                                          |
            #        |   msgbox                                                 |
            #        |                                                          |
            #        +----------------------------------------------------------+
            #

            button_group = QButtonGroup(parent=self)
            self.object_radio = object_radio = QRadioButton("Object")
            self.expr_radio = expr_radio = QRadioButton("Expression")
            button_group.addButton(object_radio)
            button_group.addButton(expr_radio)

            object_radio.toggled.connect(self.activateObject)
            expr_radio.toggled.connect(self.activateExpression)

            update_button = QPushButton(text="Update", parent=self)
            update_button.clicked.connect(self.update_data)

            # txt = _("Expression")
            # if sys.platform == 'darwin':
            #     expr_label = QLabel("  " + txt)
            # else:
            #     expr_label = QLabel(txt)

            if spyder.version_info < (4,):
                font = self.plugin.get_plugin_font()
            else:
                font = self.plugin.get_font()

            self.exprbox = MxPyExprLineEdit(self, font=font)
            self.objbox = QLabel(parent=self)
            self.argbox = MxPyExprLineEdit(self, font=font)
            self.msgbox = QLabel(parent=self)
            self.msgbox.setText("")
            self.msgbox.setWordWrap(True)

            outer_layout = QVBoxLayout()
            upper_layout = QHBoxLayout()
            outer_layout.addLayout(upper_layout)
            outer_layout.addWidget(self.msgbox)

            inner_layout = QGridLayout()
            inner_layout.addWidget(object_radio, 0, 0)
            inner_layout.addWidget(expr_radio, 1, 0)
            inner_layout.addWidget(self.exprbox, 1, 1)
            objbox_layout = QHBoxLayout()
            objbox_layout.addWidget(self.objbox)
            objbox_layout.addWidget(self.argbox)
            objbox_layout.setStretch(0, 3)  # 3:1
            objbox_layout.setStretch(1, 1)
            inner_layout.addLayout(objbox_layout, 0, 1)

            upper_layout.addLayout(inner_layout)
            upper_layout.addWidget(update_button)

            # widget = QWidget(parent=self)
            # widget.setLayout(outer_layout)

            # Create main widget
            self.widget = QWidget(parent=self) # MxDataFrameViewer(self)

            self.main_layout = main_layout = QVBoxLayout()
            main_layout.addLayout(outer_layout)
            main_layout.addWidget(self.widget)
            main_layout.setStretch(1, 1)

            # Main layout of this widget
            layout = create_plugin_layout(self.tools_layout)
            layout.addLayout(main_layout)

            margins = (5, 5, 5, 5)

            for lo in [outer_layout, upper_layout, inner_layout, self.msgbox]:
                lo.setContentsMargins(*margins)

            objbox_layout.setContentsMargins(5, 5, 0, 5)

            self.setLayout(layout)

            self.attrdict = None
            object_radio.setChecked(True)

        def setup_option_actions(self, *args, **kwargs):

            self.calc_on_update = create_action(self,
               _("Calculate upon update"),
               tip=_("Calculate the selected cells if not yet calculated"),
               toggled=lambda state:
               self.sig_option_changed.emit('calc_on_update', state))
            self.calc_on_update.setChecked(True)
            self.actions = [self.calc_on_update]

        def set_shellwidget(self, shellwidget):
            """Bind shellwidget instance to namespace browser"""
            self.shellwidget = shellwidget
            self.shellwidget.set_mxdataview(self)

        # MxToolBarMixin interface method
        def setup_toolbar(self):
            return []

        def activateObject(self, checked):
            self.argbox.setEnabled(checked)
            self.exprbox.setEnabled(not checked)

        def activateExpression(self, checked):
            self.argbox.setEnabled(not checked)
            self.exprbox.setEnabled(checked)

        def update_object(self, data):
            if data is None:
                return

            self.attrdict = data
            if data["type"] == 'Reference' or (
                    data["type"] == "Cells" and not data["parameters"]):
                self.argbox.setText("")
                self.argbox.setEnabled(False)
            else:
                self.argbox.setEnabled(True)

            self.objbox.setText(data['_evalrepr'])

        def update_data(self):
            if self.object_radio.isChecked():

                argtxt = self.argbox.get_expr()
                args = "(" + argtxt + ("," if argtxt else "") + ")"
                # assert
                ast.literal_eval(args)

                calc = self.calc_on_update.isChecked()

                val, is_calculated = self.shellwidget.update_mxdataview(
                    is_obj=True,
                    obj=self.attrdict["fullname"],
                    args=args,
                    calc=calc
                )
                self.update_value(val)
                if is_calculated:
                    self.shellwidget.refresh_namespacebrowser()

            elif self.expr_radio.isChecked():
                self.shellwidget.update_mxdataview(
                    is_obj=False,
                    expr=self.exprbox.get_expr()
                )
                self.shellwidget.refresh_namespacebrowser()
            else:
                raise RuntimeError("MxDataViewer: must not happen")

        def update_value(self, data):

            import pandas as pd
            import numpy as np
            import numpy.ma

            if isinstance(data, (pd.DataFrame, pd.Index, pd.Series)):
                self.widget.deleteLater()
                self.widget = MxDataFrameViewer(self)
                self.widget.setup_and_check(data)
                self.msgbox.setText(data.__class__.__name__)
            elif isinstance(data, (np.ndarray, np.ma.MaskedArray)):
                self.widget.deleteLater()
                self.widget = MxArrayViewer(self)
                self.widget.setup_and_check(data, title='', readonly=True)
                self.msgbox.setText(data.__class__.__name__)
            elif isinstance(data, (list, set, tuple, dict)):
                self.widget.deleteLater()
                self.widget = MxCollectionsViewer(self)
                self.widget.setup(data, title='', readonly=True)
                self.msgbox.setText(data.__class__.__name__)
            else:
                txt = data.__class__.__name__ + ": " + repr(data)
                self.widget.deleteLater()
                self.widget = QWidget(parent=self)
                self.msgbox.setText(txt)

            self.main_layout.addWidget(self.widget)
            self.main_layout.setStretchFactor(self.widget, 1)


if spyder.version_info > (5,):

    # New plugin API since Spyder 5
    from spyder.api.plugins import SpyderDockablePlugin, Plugins
    from qtpy.QtGui import QIcon
    from spyder.api.widgets.main_widget import PluginMainWidget
    # from spyder_modelx.plugins.mxplugin import ModelxConfigPage

    class MxDataViewMainWidgetActions:

        CalcOnUpdate = 'calc_on_update'

    class MxDataViewMainWidgetActionsOptionsMenuSections:

        Main = 'main_section'

    class MxDataViewMainWidget(MxStackedMixin, PluginMainWidget):

        MX_WIDGET_CLASS = MxDataViewWidget

        def __init__(self, name=None, plugin=None, parent=None):
            PluginMainWidget.__init__(self, name, plugin, parent)
            MxStackedMixin.__init__(self, parent=parent)

            # Layout
            layout = QVBoxLayout()
            layout.addWidget(self.stack)
            self.setLayout(layout)

            self.ipyconsole = plugin.get_plugin(Plugins.IPythonConsole)

        # --- API: methods to define or override
        # ------------------------------------------------------------------------
        def get_title(self):
            """
            Return the title that will be displayed on dockwidget or window title.
            """
            return _('MxDataViewer')

        def setup(self):
            """
            Create widget actions, add to menu and other setup requirements.
            """

            # ---- Toolbar actions
            self.calc_on_update_action = self.create_action(
                MxDataViewMainWidgetActions.CalcOnUpdate,
                text=_('Calculate on update'),
                tip=_('Calculate the selected if not yet calculated'),
                toggled=True
            )
            self.calc_on_update_action.setChecked(True)

            # Options menu
            options_menu = self.get_options_menu()
            for item in [self.calc_on_update_action]:
                self.add_item_to_menu(
                    item,
                    menu=options_menu,
                    section=MxDataViewMainWidgetActionsOptionsMenuSections.Main
                )

        def update_actions(self):
            """
            Update the state of exposed actions.

            Exposed actions are actions created by the self.create_action method.
            """
            pass

    class MxDataViewPlugin(SpyderDockablePlugin):
        """modelx sub-plugin.

        This plugin in registered by the modelx main plugin.
        """
        NAME = 'mxdataviewer'
        WIDGET_CLASS = MxDataViewMainWidget
        REQUIRES = [Plugins.IPythonConsole, 'modelx_plugin']
        CONF_SECTION = 'modelx'
        CONF_FILE = False

        # -------------------------------------------------------------------
        # --- API: Mandatory methods to define ------------------------------

        def get_name(self):
            """
            Return the plugin localized name.

            Returns
            -------
            str
                Localized name of the plugin.

            Notes
            -----
            This is a method to be able to update localization without a restart.
            """
            return _('MxDataViewer')

        def get_description(self):
            """
            Return the plugin localized description.

            Returns
            -------
            str
                Localized description of the plugin.

            Notes
            -----
            This is a method to be able to update localization without a restart.
            """
            return _('Widget for tracing modelx node dependency')

        def get_icon(self):
            """
            Return the plugin associated icon.

            Returns
            -------
            QIcon
                QIcon instance
            """
            return QIcon()

        def register(self):
            """
            Setup and register plugin in Spyder's main window and connect it to
            other plugins.
            """
            self.get_plugin('modelx_plugin').get_container().set_child_plugin('dataview', self.get_container())
        # -------------------------------------------------------------------

else:
    class MxDataViewPlugin(MxStackedMixin, SpyderPluginWidget):
        """modelx sub-plugin.

        This plugin in registered by the modelx main plugin.
        """

        CONF_SECTION = 'modelx_dataviewer'
        MX_WIDGET_CLASS = MxDataViewWidget
        CONF_FILE = False

        def __init__(self, parent=None, **kwargs):

            SpyderPluginWidget.__init__(self, parent)
            MxStackedMixin.__init__(self, parent)

            # Layout
            layout = QVBoxLayout()
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(self.stack)
            if spyder.version_info > (4,):
                self.options_button.setVisible(False)
            self.setLayout(layout)

            if spyder.version_info < (4,):
                # Initialize plugin
                self.initialize_plugin()

        # --- SpyderPluginWidget API ----------------------------------------------
        def get_plugin_title(self):
            """Return widget title."""
            return 'MxDataViewer'

        def get_focus_widget(self):
            """Return the widget to give focus to."""
            return self.current_widget()

        def refresh_plugin(self):
            """Refresh MxExplorer widget."""
            pass

        def get_plugin_actions(self):
            """Return a list of actions related to plugin"""
            widget = self.current_widget()
            if isinstance(widget, self.MX_WIDGET_CLASS):
                return widget.actions
            else:
                # Start-up Label
                return []

        def register_plugin(self):
            """Register plugin in Spyder's main window."""
            if spyder.version_info < (4,):
                self.main.add_dockwidget(self)
            else:
                self.add_dockwidget()

        def on_first_registration(self):
            """Action to be performed on first plugin registration."""
            self.main.tabify_plugins(self.main.help, self)

        def apply_plugin_settings(self, options):
            """Apply configuration file's plugin settings."""
            pass
